from collections.abc import Sequence
from dataclasses import dataclass

import pytest

from memium.utils.get_wikilinks_from_string import get_wikilinks_from_string


@dataclass(frozen=True)
class TestExample:
    input_string: str
    wikilinks: Sequence[str]


@pytest.mark.parametrize(
    ("example"),
    [
        TestExample("[[Test|Test2]]", ["Test"]),
        TestExample("[[Test (testing)]]", ["Test (testing)"]),
        TestExample("[[Test]]", ["Test"]),
        TestExample("[[Test]], [[Test2]]", ["Test", "Test2"]),
        TestExample("[[ Some things, are just not [[Wikilinks]] ]]", ["Wikilinks"]),
    ],
    ids=lambda x: x.input_string,
)
def test_get_wikilinks_from_string(example: TestExample):
    assert get_wikilinks_from_string(example.input_string) == example.wikilinks



import enum
import re
from collections.abc import Sequence
from dataclasses import dataclass
from typing import Literal

from memium.source.document import Document
from memium.source.prompts.prompt import BasePrompt

from ..prompts.prompt_qa import QAPrompt
from .extractor import BasePromptExtractor


class TableParseMode(enum.Enum):
    ASCENDING = "Ascending"
    DESCENDING = "Descending"
    ROWWISE = "Rowwise"


@dataclass(frozen=True)
class ParsedTable:
    rows: list[dict[str, str]]
    mode: TableParseMode
    front: str
    back: str


@dataclass(frozen=True)
class RowPair:
    front: dict[str, str]
    back: dict[str, str]


class TableExtractor(BasePromptExtractor):
    def _parse_table(self, input_str: str) -> Sequence[ParsedTable]:
        if not input_str.startswith("|"):
            return []

        rows: list[dict[str, str]] = []
        parsed_tables: Sequence[ParsedTable] = []
        for n, line in enumerate(input_str.strip().split("\n")):
            data: dict[str, str] = {}
            if n == 0:
                header = [cell.strip() for cell in line.split("|") if cell]
            if n > 1:
                # Is a row
                if line.startswith("|"):
                    values = [cell.strip() for cell in line.split("|") if cell]
                    for col, value in zip(header, values, strict=False):  # type: ignore
                        data[col] = value
                    rows.append(data)
                # Is description
                elif "//" in line:
                    metadata = line.split("//")
                    mode_string: Literal["Ascending", "Descending", "Rowwise", "Row-wise"] = (
                        metadata[0].strip()
                    )  # type: ignore

                    match mode_string:
                        case "Ascending":
                            mode = TableParseMode.ASCENDING
                        case "Descending":
                            mode = TableParseMode.DESCENDING
                        case "Rowwise" | "Row-wise":
                            mode = TableParseMode.ROWWISE

                    parsed_tables.append(
                        ParsedTable(
                            rows=rows,
                            mode=mode,
                            front=metadata[1].strip(),
                            back=metadata[2].strip(),
                        )
                    )
        return parsed_tables

    def _get_row_pair(self, parsed_table: ParsedTable, i: int) -> RowPair:
        match parsed_table.mode:
            case TableParseMode.ASCENDING:
                front_row = parsed_table.rows[-1 - i]
                back_row = parsed_table.rows[-2 - i]
            case TableParseMode.DESCENDING:
                front_row = parsed_table.rows[i]
                back_row = parsed_table.rows[i + 1]
            case TableParseMode.ROWWISE:
                front_row = parsed_table.rows[i]
                back_row = parsed_table.rows[i]

        return RowPair(front_row, back_row)

    def _replace_placeholders(
        self, parsed_table: ParsedTable, row_pair: RowPair
    ) -> tuple[str, str]:
        front_placeholders = re.findall(r"\|(.+?)\|", parsed_table.front)
        back_placeholders = re.findall(r"\|(.+?)\|", parsed_table.back)
        front = parsed_table.front
        back = parsed_table.back

        for placeholder in [*front_placeholders, *back_placeholders]:
            front = front.replace(f"|{placeholder}|", row_pair.front[placeholder])
            back = back.replace(f"|{placeholder}|", row_pair.back[placeholder])

        return front, back

    def _parsed_table_to_prompt(self, parsed_table: ParsedTable) -> Sequence[QAPrompt]:
        prompts: Sequence[QAPrompt] = []
        match parsed_table.mode:
            case TableParseMode.ASCENDING | TableParseMode.DESCENDING:
                break_index = len(parsed_table.rows) - 2
            case TableParseMode.ROWWISE:
                break_index = len(parsed_table.rows) - 1

        for i in range(len(parsed_table.rows)):
            row_pair = self._get_row_pair(parsed_table, i)

            front, back = self._replace_placeholders(parsed_table, row_pair)

            # Skip rows with empty fronts or backs
            if front and back:
                prompts.append(QAPrompt(question=front, answer=back))
            if i == break_index:
                break
        return prompts

    def extract_prompts(self, document: Document) -> Sequence[BasePrompt]:
        blocks = document.content.split("\n\n")
        parsed_tables = [self._parse_table(block) for block in blocks]
        flattened_parsed_tables = [
            parsed_table for sublist in parsed_tables for parsed_table in sublist
        ]
        prompts = [
            self._parsed_table_to_prompt(parsed_table) for parsed_table in flattened_parsed_tables
        ]
        flattened_prompts = [prompt for sublist in prompts for prompt in sublist]
        return flattened_prompts



import re
from collections.abc import Sequence


def get_wikilinks_from_string(string: str) -> Sequence[str]:
    wikilinks = re.findall(r"\[\[[\w\s\|\(\)]+\]\]", string)
    without_alias = [re.sub(r"\|[\w|\s|\d]+\]\]", "]]", link) for link in wikilinks]
    contents = [link.replace("[[", "").replace("]]", "") for link in without_alias]
    return contents



import hashlib
import logging
import re
from collections.abc import Sequence

from ..document import Document
from ..prompts.prompt_cloze import ClozeFromDoc, ClozePrompt
from .extractor import BasePromptExtractor

log = logging.getLogger(__name__)


class ClozePromptExtractor(BasePromptExtractor):
    @staticmethod
    def _get_blocks(string: str) -> list[str]:
        """Break string into a list by 2+ newlines in a row."""
        # Exclude entire code blocks
        string_sans_code_blocks = re.sub(r"```.*?```", "", string, flags=re.DOTALL)
        return re.split(r"(\n\n)+", string_sans_code_blocks)

    @staticmethod
    def _has_cloze(string: str) -> bool:
        if len(re.findall(r"{.*}", string)) > 0:
            return True
        return False

    @staticmethod
    def _is_math_block(string: str) -> bool:
        if string.startswith("$$"):
            return True
        return False

    @staticmethod
    def _is_html_comment(string: str) -> bool:
        if string.startswith("<!--"):
            return True
        return False

    @staticmethod
    def _replace_cloze_id_with_unique(string: str, selected_cloze: str | None = None) -> str:
        """Each cloze deletion in a note is numbered sequentially.

        This function ensures that the numbering is based on the content of the cloze deletion, essentially ensuring that if you modify the contents of a cloze, only the scheduling of that specific cloze is changed.

        Args:
            string (str): The string to replace the cloze id with a unique id.
            selected_cloze (str, optional): If you only want to replace a specific cloze, pass it here. Defaults to None.
        """
        if selected_cloze is not None:
            selected_clozes = [selected_cloze]
        else:
            selected_clozes = re.findall(r"{(?!BearID).[^}]*}", string)

        for cloze in selected_clozes:
            output_hash = int(hashlib.sha256(cloze.encode("utf-8")).hexdigest(), 16) % 10**3

            new_cloze = f"{{{{c{output_hash}::{cloze[1:-1]}}}}}"

            string = string.replace(cloze, new_cloze)

        return string

    def extract_prompts(self, document: Document) -> Sequence[ClozePrompt]:
        prompts: list[ClozeFromDoc] = []
        blocks = self._get_blocks(document.content)

        block_starting_line_nr = 1
        for block_string in blocks:
            if any(
                exclusion_criterion(block_string)
                for exclusion_criterion in (self._is_html_comment, self._is_math_block)
            ):
                continue

            if self._has_cloze(block_string):
                clozes = re.findall(r"{(?!BearID).[^}]*}", block_string)

                for selected_cloze in clozes:
                    prompt_content = self._replace_cloze_id_with_unique(
                        block_string, selected_cloze=selected_cloze
                    )

                    prompts.append(
                        ClozeFromDoc(
                            text=prompt_content, parent_doc=document, line_nr=block_starting_line_nr
                        )
                    )

            n_block_lines = len(re.findall(r"\n", block_string, flags=re.DOTALL))
            block_starting_line_nr += n_block_lines

        return prompts



import logging
import time
from functools import partial
from pathlib import Path
from typing import Annotated, Optional

import sentry_sdk
import typer

from memium.destination.ankiconnect.anki_converter import AnkiPromptConverter
from memium.destination.ankiconnect.ankiconnect_gateway import ANKICONNECT_URL, AnkiConnectGateway
from memium.destination.destination import PushPrompts
from memium.destination.destination_ankiconnect import AnkiConnectDestination
from memium.destination.destination_dryrun import DryRunDestination
from memium.diff_determiner import PromptDiffDeterminer
from memium.environment import get_env, host_input_dir, in_docker
from memium.source.document_source import MarkdownDocumentSource
from memium.source.extractors.extractor_qa import QAPromptExtractor
from memium.source.prompt_source import DocumentPromptSource

from .source.extractors.extractor_table import TableExtractor

log = logging.getLogger(__name__)

app = typer.Typer()


def main(
    base_deck: str,
    input_dir: Path,
    max_deletions_per_run: int,
    dry_run: bool,
    push_all: bool = False,
):
    source_prompts = DocumentPromptSource(
        document_ingester=MarkdownDocumentSource(directory=input_dir),
        prompt_extractors=[
            QAPromptExtractor(question_prefix="Q.", answer_prefix="A."),
            TableExtractor(),
        ],
    ).get_prompts()

    dest_class = AnkiConnectDestination if not dry_run else DryRunDestination
    destination = dest_class(
        gateway=AnkiConnectGateway(
            ankiconnect_url=ANKICONNECT_URL,
            base_deck=base_deck,
            tmp_read_dir=host_input_dir() if in_docker() else input_dir,
            tmp_write_dir=input_dir,
            max_deletions_per_run=max_deletions_per_run,
            max_wait_seconds=30,
        ),
        prompt_converter=AnkiPromptConverter(
            base_deck=base_deck,
            card_css=Path("memium/destination/ankiconnect/default_styling.css").read_text(),
        ),
    )
    if not push_all:
        destination_prompts = destination.get_all_prompts()
        update_commands = PromptDiffDeterminer().sync(
            source_prompts=source_prompts, destination_prompts=destination_prompts
        )
        destination.update(commands=update_commands)
    else:
        destination.update([PushPrompts(prompts=source_prompts)])


@app.command()
def cli(
    input_dir: Annotated[
        Path,
        typer.Option(
            help="Directory containing markdown prompts to sync to Anki.",
            dir_okay=True,
            file_okay=False,
            exists=True,
            readable=True,
            writable=True,
        ),
    ],
    watch_seconds: Annotated[
        Optional[int],  # noqa: UP007
        typer.Option(help="Keep running, updating Anki deck every [ARG] seconds"),
    ] = None,
    deck_name: Annotated[
        str, typer.Option(help="Anki path to deck, e.g. 'Parent deck::Child deck'")
    ] = "Memium",
    max_deletions_per_run: Annotated[
        int,
        typer.Option(
            help="Maximum number of cards to delete per sync to avoid unintentional deletions. If exceeded, raises error."
        ),
    ] = 50,
    push_all: Annotated[
        bool,
        typer.Option(
            help="Push all prompts to Anki, whether or not they exist in Anki. Useful if you have e.g. changed your CSS template. Note that this does not change scheduling, nor delete prompts that no longer exist in your markdown."
        ),
    ] = False,
    dry_run: Annotated[
        bool, typer.Option(help="Don't update via AnkiConnect, just log what would happen")
    ] = False,
    sentry_dsn: Annotated[
        Optional[str],  # noqa: UP007
        typer.Option(
            help="Sentry DSN for error reporting. If not provided, no error reporting will be done."
        ),
    ] = None,
    skip_sync: Annotated[bool, typer.Option(help="Skip all syncing, useful for smoketest")] = False,
):
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s [%(levelname)s] %(name)s: %(message)s",
        datefmt="%Y/&m/%d %H:%M:%S",
    )

    if sentry_dsn:
        sentry_sdk.init(dsn=sentry_dsn, environment=get_env(default="None"))

    main_fn = partial(
        main,
        base_deck=deck_name,
        input_dir=input_dir,
        max_deletions_per_run=max_deletions_per_run,
        dry_run=dry_run,
        push_all=push_all,
    )

    if not skip_sync:
        main_fn()
        if watch_seconds:
            log.info(f"Sync complete, sleeping for {watch_seconds} seconds")
            time.sleep(watch_seconds)
            main_fn()


if __name__ == "__main__":
    app()
